;;;
;;; Copyright (C) 2010 Frederic Vecoven
;;;
;;; This file is part of NixieClock
;;;
;;; NixieClock is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; NixieClock is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program. If not, see <http://www.gnu.org/licenses/>.
;;;
	
;
; Clock routines
;
	include hv.inc
	include program.inc
	
	; variables
	extern	GFlag, TFlag, PFlag, CFlag, sfsr, rand_tmp, max_bright
	extern	time, tick, pair0, pair1, tcount, tdigit, qsec
	extern	auto_values, auto_brightness, auto_fading, auto_flags
	extern	drift, drift_value, fade_tmp
	
	; functions
	extern	set_fading, get_tube, set_tube, freq_mask, fading_table
	extern	get_bright, set_bright
	extern	update_qsec_ptr, update_minute_ptr, update_hour_ptr
	extern	update_day_ptr, update_week_ptr, rand_10, handle_buttons
	extern	incr_time, jump_table, hours24_to_12

#if SUPPORT_IR == 1
	; variable
	extern	ir_timeout
	
	; function
	extern	ir_check_timeout
#endif

#if SUPPORT_DP == 1
	; variable
	extern	auto_dp

	; function
	extern	set_dp
#endif
	
	
	
clock_code	CODE


;
; Timer 1 : real-time clock
;
; This function is called 64 times per second.
;
; The oscillator runs at 32.768kHz, therefore 1 second = 0x8000 ticks. 
; To get 64 interrupts per second, we need 0x8000 / 64 = 0x200 ticks,
; and the reload value is therefore 0x10000 - 0x200 = 0xFE00.
;
; We can compensate a crystal running too slowly or too quickly by
; correcting the reload value from time to time. To do this, a 16 bits
; value "drift" is used. Every 64th second, the value is decremented. When 
; it reaches 0, the timer reload value is altered to make the crystal appear
; running slower or faster.
;
; For example, to compensate for -2 seconds per day :
;    1) number of 64th seconds per day = 24 * 60 * 60 * 64
;    2) 2 seconds = 32768 ticks * 2 = 65536 
;    3) solve   24*60*60*64/(65536/drift_value) == drift
;       ==> drift_value = 8, drift = 675
;
int_timer1
	GLOBAL	int_timer1
	bcf	PIR1,TMR1IF
	
#if 0
	decf	drift+2,f		; decrement drift (16 bits)
	btfss	STATUS,C
	decf	drift+3,f
	bnz	drift_nz		; test if drift == 0
	tstfsz	drift+2
	bra	drift_nz
drift_z
	movff	drift+0,drift+2		; reload drift
	movff	drift+1,drift+3
	movlw	0xFE			; reload timer 1
	btfsc	drift_value,7		;    drift value == 0  ==> 0xFE00
	decf	WREG,w			;    drift value == +x ==> 0xFE..
	movwf	TMR1H			;    drift value == -x ==> 0xFD..
	movf	drift_value,w
	movwf	TMR1L
	bra	drift_done
#endif

drift_nz
	movlw	0xFE			; 0x200 = 0x10000 - 0xFE00 (1/64th sec)
	movwf	TMR1H			; (1 second = 0x8000)
	clrf	TMR1L
drift_done
					; increment tick counter
	incf	tick,f

#if SUPPORT_IR == 1			; decrement IR timeout
	tstfsz	ir_timeout
	call	ir_check_timeout
#endif	
					; call button handler
	call	handle_buttons
					; every 64 ticks, one second has elapsed
	movf	tick,w
	andlw	0x3F
	btfsc	STATUS,Z
	call	incr_time
					; main loop to handle remaining work
	bsf	GFlag,TIMER1_WORK

	return


;
; Timer1 work : update the tube values (every 64th second), and
;               the program pointers (every 16th second).
;
; This is called from the main loop, if timer1 has set a flag.
;	
; Note that this routine may be interrupted by the tube refresh interrupt
; (running from timer0). This interrupt routine rebuilds its data structure
; when TUBE_UPDATE is set. So, before updating the values used by the
; build function in timer0, we disable the low priority interrupt if the
; TUBE_UPDATE flag is still set. This avoids timer0 from using data being
; currently updated by this routine.
;
timer1_work
	GLOBAL	timer1_work

	btfss	GFlag,TIMER1_WORK	; something to do ?
	return
					; yes. acknowledge and do it...
	bcf	GFlag,TIMER1_WORK
					; if power is lost, we are done
	btfsc	PFlag,POWER_LOST
	return
					; skip the program in manual mode	
	btfsc	GFlag,MANUAL_UPDATE
	bra	tw_10
	
	btfsc	TFlag,E_WEEK
	call	update_week_ptr
	btfsc	TFlag,E_DAY
	call	update_day_ptr
	btfsc	TFlag,E_HOUR
	call	update_hour_ptr
	btfsc	TFlag,E_MIN
	call	update_minute_ptr
	call	update_qsec_ptr
					; if tubes are off, we are done
	btfss	PFlag,POWER_ON
	return
tw_10					; every tick (1/64th sec), update tubes
	btfsc	GFlag,TUBE_UPDATE
	bcf	INTCON,GIEL
	call	update_values
	call	update_fading
	call	update_brightness
#if SUPPORT_DP == 1
	call	update_dp
#endif
	bsf	INTCON,GIEL
					; force a recomputation of tube data
	bsf	GFlag,TUBE_UPDATE

	return	

;
; auto-brightness
;
; This routine is called every 1/64 second, and is reponsible to
; update the brightness of each tube.
;
; Entries are ordered like this : TUBE1 TUBE2 TUBE3 ...
;
; The 3 MSB indicate what to do :
;
; 0 0 0  x x x x x - do nothing
; 0 0 1  x x F F F - increment. 3 LSB = frequency
; 0 1 0  x x F F F - decrement. 3 LSB = frequency
; 0 1 1  x x F F F - increment/decrement cycle. 3 LSB = frequency
; 1 0 0  x D D D D - direct value
;
update_brightness
	lfsr	FSR1,auto_brightness
	clrf	tcount
ub_00
	swapf	INDF1,w			; get 3 MSB
	rrcf	WREG,w
	andlw	0x07			; into W
	call	jump_table
	goto	ub_next			; 0 = do nothing
	goto	bright_inc		; 1 = increment
	goto	bright_dec		; 2 = decrement
	goto	bright_cycle		; 3 = cycle
	goto	bright_direct		; 4 = set direct
	goto	ub_next
	goto	ub_next
	goto	ub_next
ub_next
	movf	POSTINC1,w		; FSR1++
	incf	tcount,f		; tcount++
	movlw	NUM_TUBES
	cpfseq	tcount			; last tube ?
	bra	ub_00
	return
bright_inc
	movf	INDF1,w
	call	skip_freq
	bnz	ub_next
	call	get_bright
	incf	WREG,w
	call	set_bright
	bra	ub_next
bright_dec
	movf	INDF1,w
	call	skip_freq
	bnz	ub_next
	call	get_bright
	decf	WREG,w
	call	set_bright
	bra	ub_next
bright_cycle
	movf	INDF1,w
	call	skip_freq
	bnz	ub_next
	call	get_bright
	movwf	tdigit
	lfsr	FSR2,auto_flags
	movf	tcount,w
	addwf	FSR2L,f
	btfsc	INDF2,BRIGHT_CYCLE
	bra	cycle_dec
	movf	max_bright,w
	cpfseq	tdigit
	bra	cycle_inc
	bsf	INDF2,BRIGHT_CYCLE
	bra	cycle_dec
cycle_inc
	incf	tdigit,w
	call	set_bright
	bra	ub_next
cycle_dec
	movlw	BRIGHT_MIN
	cpfsgt	tdigit
	bra	cycle_to_inc
	decf	tdigit,w
	call	set_bright
	bra	ub_next
cycle_to_inc
	bcf	INDF2,BRIGHT_CYCLE
	bra	cycle_inc
	
bright_direct
	movf	INDF1,w
	call	set_bright
	bra	ub_next
	


;
; auto-values
;
; This routine is called every 1/64th second, and is responsible to update the
; digit of each tube.
;
; Entries are ordered like this : TUBE1, TUBE1_ALT, TUBE2, TUBE2_ALT, etc...
;
; The 4 MSB indicate what to do :
;
; 0 0 0 0  x x x x - do nothing
; 0 0 0 1  x F F F - increment. The 3 LSB give the frequency
; 0 0 1 0  x F F F - decrement. The 3 LSB give the frequency
; 0 0 1 1  x F F F - random. The 3 LSB give the frequency
; 0 1 0 0  D D D D - direct value, found in 4 LSB.
;	
; 1 S O O  O O O N - auto set. If S is 0, then the bits [5..1] give the offset
;                    in 'time'. If S is 1, then bits [5..1] maps to a dedicated
;                    function which sets the value according to option flags.
;                    bit 0 always indicates which nibble we want.
;
update_values
	lfsr	FSR1,auto_values
	clrf	tcount
uv_00
	btfsc	INDF1,7			; bit 7 set ?
	bra	value_autoset		; yes... auto set
	swapf	INDF1,w			; no... get 4 MSB (0x0 -> 0x7 since 
	andlw	0x0F			; bit 7 is 0) into W
	call	jump_table
	goto	uv_next			; 0 = do nothing
	goto	value_inc		; 1 = increment
	goto	value_dec		; 2 = decrement
	goto	value_random		; 3 = random
	goto	value_set		; 4 = set direct
	goto	uv_next			; 5 = not used
	goto	uv_next			; 6 = not used
	goto	uv_next			; 7 = not used
uv_next		
	movf	POSTINC1,w		; FSR1++
	incf	tcount,f		; tcount++
	movlw	NUM_TUBES*2		; (x2 for alternate values)
	cpfseq	tcount			; last tube ?
	bra	uv_00
	return

value_autoset
	rrcf	INDF1,w			; get bits [6..1] into W[5..0]
	andlw	0x3F
	btfsc	WREG,5			; value depends on options ?
	bra	autoset_options		; yes.. handle it
	lfsr	FSR2,t_sec		; no... just copy. FSR2 : ptr to time
	movf	PLUSW2,w		; PLUSW2 = time[bits[5..1]]
	btfsc	INDF1,0			; if bits[0] == '1'
	swapf	WREG,w			; then take the other nibble
	call	set_tube
	bra	uv_next
autoset_options
	lfsr	FSR2,t_sec
	andlw	0x7			; keep lower 3 bits (for now)
	call	jump_table
	goto	autoset_date_part1
	goto	autoset_date_part2
	goto	autoset_date_part1_1
	goto	autoset_date_part2_1
	goto	autoset_hours
	goto	autoset_hours_1
	goto	uv_next
	goto	uv_next

autoset_date_part1
	movlw	(DAY_H >> 1)		; first part default = DAY
	btfsc	CFlag,DATE_MMDD_BIT
	movlw	(MONTH_H >> 1)		; if MMDD is set, then it's MONTH
daset	movf	PLUSW2,w
	btfsc	INDF1,0
	swapf	WREG,w
	call	set_tube
	bra	uv_next
autoset_date_part2
	movlw	(MONTH_H >> 1)		; second part default = MONTH
	btfsc	CFlag,DATE_MMDD_BIT
	movlw	(DAY_H >> 1)		; if MMDD is set, then it's DAY
	bra	daset
autoset_date_part1_1
	movlw	(DAY_1_H >> 1)
	btfsc	CFlag,DATE_MMDD_BIT
	movlw	(MONTH_1_H >> 1)
	bra	daset
autoset_date_part2_1
	movlw	(MONTH_1_H >> 1)
	btfsc	CFlag,DATE_MMDD_BIT
	movlw	(DAY_1_H >> 1)
	bra	daset
	
autoset_hours
	movlw	(HOURS24_H >> 1)	; read hours
autoset_hours_00
	movf	PLUSW2,w
	btfsc	CFlag,MODE12_BIT	; if (12-hours mode) then
	call	hours24_to_12		;    convert hours
	btfsc	CFlag,ZERO_SUPPRESS_BIT	; if (suppress leading 0) then
	call	suppress_0		;    supress zero
	btfsc	INDF1,0			; select the correct nibble
	swapf	WREG,w
	call	set_tube
	bra	uv_next
autoset_hours_1
	movlw	(HOURS24_1_H >> 1)	; read hours in time+1sec
	bra	autoset_hours_00
	
value_inc
	movf	INDF1,w			; handle frequency setting
	call	skip_freq
	bnz	uv_next
	call	get_tube		; incr current digit
	incf	WREG,w
	daw
	call	set_tube
	bra	uv_next

value_dec
	movf	INDF1,w			; handle frequency setting
	call	skip_freq
	bnz	uv_next
	call	get_tube		; decr current digit
	decf	WREG,w
	btfsc	STATUS,N
	movlw	0x09
	call	set_tube
	bra	uv_next

value_random
	movf	INDF1,w			; handle frequency setting
	call	skip_freq
	bnz	uv_next
	call	get_tube
	movwf	rand_tmp		; save previous tube value
vr_rand	call	rand_10
	cpfseq	rand_tmp		; if same
	bra	vr_set
	bra	vr_rand			; then get another random value	
vr_set	call	set_tube
	bra	uv_next
value_set
	movf	INDF1,w			; direct set value
	andlw	0x0F
	call	set_tube
	bra	uv_next

; suppress a leading 0 in W : if W's high nibble is 0, it gets converted
; to F. 
suppress_0
	movwf	rand_tmp		; save in temp variable
	andlw	0xF0			; if (W & 0xF0 == 0)
	bz	sup0
	movf	rand_tmp,w
	return
sup0	movf	rand_tmp,w
	iorlw	0xF0			; then W |= 0xF0
	return
	

;
; skip_freq 
;
; set zero flag if the operation should be conducted at the given freq (in W).
;
; W = 0  : every 2 sec 
;     1          1 sec
;     2          1/2 sec
;     3          1/4 sec
;     4          1/8 sec
;     5          1/16 sec
;     6          1/32 sec
;     7          1/64 sec
skip_freq
	call	freq_mask
	andwf	tick,w
	return
	

;
; auto-fading
;
; This routine is called every 1/64th second, and is responsible to update the
; various fading values for each tube.
;
;
; The 3 MSB indicate what to do :
;
; 0 0 0  x x x x x - no fading
; 0 0 1  x F F F F - no fading for 15-F 16th sec, than fade during F+1 16th sec
; 0 1 0  x x x F F - repeat fading (0=2x, 1=4x, 2=8x, 3=16x per second)
; 1 1 1  D D D D D - direct value, expressed in 1/32 sec
;
update_fading
	lfsr	FSR1, auto_fading
	clrf	tcount
af_00
	swapf	INDF1,w			; get 3 MSB
	rrcf	WREG,w
	andlw	0x07			; into W
	call	jump_table
	goto	no_fading		; 0
	goto	fading_type_1		; 1
	goto	fading_type_2		; 2
	goto	no_fading		; 3
	goto	no_fading		; 4
	goto	no_fading		; 5
	goto	no_fading		; 6
	goto	fading_direct		; 7

af_next
	movf	POSTINC1,w		; FSR1++
	incf	tcount,f		; tcount++
	movlw	NUM_TUBES
	cpfseq	tcount			; if (tcount != NUM_TUBES)
	bra	af_00			; then loop
	return


;;;
;;; fading type 1
;;;
;;; This is the typical fading : the main digit remains for a period
;;; of time, then it fades out while the alternate digit is fading in.
;;;
;;; There are 16 different fading speeds. Speed is specified in 16th
;;; of second. If speed is 0, it means that the alternate digit will
;;; start showing up after 15/16sec, and will fade in in 1/16sec.
;;; If speed is 15, it means that the alternate digit will immediately
;;; fade in, and the complete fade process will take 1 second.
;;;
;;; The algorithm is quite simple. Input parameters are
;;;
;;; 	speed : 0 - F
;;;     ticks : 0 - FF  (64 ticks per second)
;;;
;;;   x = 64 - 4 * (speed + 1)
;;;   m = 1024 / (64 - x)
;;;   if (m > 255) m = 255
;;;   a = (tick & 0x3F) - x;
;;;   if (a < 0) then
;;;        fading = 0
;;;   else
;;;        fading = (a * m) >> 5;
;;;
;;;
;;;  Note: x is computed each time, but m values are precomputed in a table.
;;; 
fading_type_1
	movf	INDF1,w			; get fading speed
	andlw	0xF			; into W
	call	fading_table		; get precomputed factor
	movwf	fade_tmp+1
	movlw	0x40			; 64
	movwf	fade_tmp
	movf	INDF1,w			; get fading speed again
	andlw	0xF
	incf	WREG,f			; W 0..15 is now 1..16
	addwf	WREG,f
	addwf	WREG,f
	subwf	fade_tmp,f		; fade_tmp = 64 - speed * 4
	negf	fade_tmp		; negate it

	movf	tick,w
	andlw	0x3f
	addwf	fade_tmp,w
	bn	no_fading
	mulwf	fade_tmp+1
	rlcf	PRODL,f			; to get PROD >> 5, shift left 3 times
	rlcf	PRODH,f			; and keep PRODH.
	rlcf	PRODL,f
	rlcf	PRODH,f
	rlcf	PRODL,f
	rlcf	PRODH,f
	movf	PRODH,w
	call	set_fading
	bra	af_next


;;;
;;; fading type 2
;;;
;;; Fading is repeated multiple times per second. This is useful to have
;;; fading during auto increment/decrement.
;;;
;;; The 2 LSB provides the rate :
;;;
;;; 0 = 2x per second
;;; 1 = 4x per second
;;; 2 = 8x per second
;;; 3 = 16x per second
;;; 
fading_type_2
	movff	tick,fade_tmp		; copy tick
	movf	INDF1,w			; get fading repeat rate
	andlw	0x3			; (only 2 bits)
	bz	ft2_01
ft2_00	rlcf	fade_tmp,f		; fade_tmp <<= repeat_rate
	decfsz	WREG,f
	bra	ft2_00
ft2_01	movf	fade_tmp,w
	andlw	0x1f			; get 5 lower bits
	call	set_fading
	bra	af_next


no_fading				; no fading : main digit only
	clrf	WREG
	call	set_fading
	bra	af_next

fading_direct				; direct value (0-31)
	movf	INDF1,w
	call	set_fading
	bra	af_next


;
; auto-dp
;
; This routine is called every 1/64 second, and is reponsible to
; update the decimal dot of each tube
;
; Entries are ordered like this : TUBE1 TUBE2 TUBE3 ...
;
; The 2 MSB indicate what to do :
;
; 0 0  x x x x x x - do nothing
; 0 1  A A A B B B - wait B 1/8s then ON for A 1/8s
; 1 0  
; 1 1  x x x x x D - direct value (0 = off, 1 = on, 2 = toggle)
;
#if SUPPORT_DP == 1
update_dp
	lfsr	FSR1,auto_dp
	clrf	tcount
udp_00
	swapf	INDF1,w			; get 2 MSB
	rrcf	WREG,w
	rrcf	WREG,w
	andlw	0x3
	call	jump_table
	goto	udp_next		; 0 = do nothing
	goto	udp_type1		; 1 = type 1
	goto	udp_next
	goto	udp_direct		; 3 = set direct
udp_next
	movf	POSTINC1,w		; FSR1++
	incf	tcount,f		; tcount++
	movlw	NUM_TUBES
	cpfseq	tcount			; last tube ?
	bra	udp_00
	return
udp_direct
	movf	INDF1,w
	call	set_dp
	bra	udp_next
udp_type1
;;;
;;; algorithm:
;;;
;;; if (duration == 0) return 0
;;; start = delay * 8
;;; stop = start + duration * 8
;;; if ((tick & 0x3f) < start || (tick & 0x3f) >= stop) return 0
;;; return 1
;;;
	movf	tick,w
	andlw	0x3f
	movwf	fade_tmp		; fade_tmp = tick & 0x3f
	movf	INDF1,w
	andlw	0x38			; get duration (already << 3)
	movwf	fade_tmp+1
	bz	udp1_set0		; if duration == 0, set 0
	rlcf	INDF1,w
	rlcf	WREG,w
	rlcf	WREG,w
	andlw	0x38			; W = delay * 8
	cpfslt	fade_tmp		; (tick & 0x3f) < start ?
	bra	udp1_00			; no.
udp1_set0				; yes. set 0
	clrf	WREG
	call	set_dp
	bra	udp_next
udp1_00
	addwf	fade_tmp+1,w		; W = stop = (delay + duration) << 3
	cpfslt	fade_tmp		; (tick & 0x3f) < stop ?
	bra	udp1_set0		; no, set 0
udp1_set1				; yes. set 1
	movlw	0x1
	call	set_dp
	bra	udp_next
#endif
	

	END
