;;;
;;; Copyright (C) 2010 Frederic Vecoven
;;;
;;; This file is part of NixieClock
;;;
;;; NixieClock is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; NixieClock is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program. If not, see <http://www.gnu.org/licenses/>.
;;;

	include hv.inc
	include	program.inc

#if SUPPORT_IR == 1
	
	; variables
	extern	ir_t0, ir_t1, ir_len, ir_timeout, ir_data, ir_bitcount
	extern	ir_state, GFlag

	
; states
#define IR_9MS		0	; 1 = we got the 9ms pulse
#define IR_4MS5		1	; 1 = we got the 4.5ms pulse
#define IR_560US	2	; 1 = we got a 560us pulse (start of 0 or 1)
#define IR_GOT_FIRST	3	; 1 = we got the first edge
#define IR_T1_ROLLOVER	4	; 1 = timer1 got reloaded

; values (multiple of 1/32768sec = ~30us)
#define T_9MS		.294
#define T_4MS5		.147
#define T_560US		.18
#define T_1MS680	.55	; 2.24ms - 560us
#define MARGIN		.50	; in %
	
	
ir_code	CODE


;
; Stop IR reception. This is called when a valid command has been received.
; The user must re-enable IR reception by calling 'init_ir'.
;
stop_ir
	bcf	IR_INT_IE
	bcf	IR_INT_IF
	return

	
;
; check_timeout : called every 1/64s if ir_timeout is not zero
;
ir_check_timeout
	GLOBAL	ir_check_timeout
	bsf	ir_state,IR_T1_ROLLOVER	; timer1 has rolled over
	dcfsnz	ir_timeout,f		; decrement timeout
	bra	init_ir			; if 0, reset IR.
	return				; otherwise, return

	
;
; IR interrupt, when IR_PIN is toggled
;
; IR pulses are measured using timer1. Timer1 runs at 32.768kHz, therefore
; one tick is ~30us. timer1 rolls over at 0x200 (1/64sec), so we have to
; use 16 bit values. If timer1 rolls over twice, we have a timeout and
; we reset the IR routines. Otherwise, we compute the difference between
; the two samples of TMR1 in order to measure the pulse in 30us steps.
; We disable high priority interrupts while doing this, so we get an
; accurate view of TMR1 and the rollover flag. This won't affect the clock
; accuracy because we have 244 instructions before we miss a pulse on
; timer1.
;	
int_ir
	GLOBAL	int_ir
	
	bcf	IR_INT_IF

	btfsc	IR_EDGE_REG,IR_EDGE_BIT	; waiting for a low edge ?
	bra	ir_wait_hi		; no. 
					; yes. filter short pulse caused by multiplex
	btfss	IR_PIN			; pin low ?
	bra	ir_tog			; yes. real pulse
	return				; no. it was a spurious. ignore it.
ir_wait_hi
	btfss	IR_PIN			; pin hi
	return				; no. it was a spurious. ignore it.					
ir_tog
	clrf	ir_timeout
	btg	IR_EDGE_REG,IR_EDGE_BIT	; reverse edge 

	btfsc	ir_state,IR_GOT_FIRST	; is this the very first edge ?
	bra	ir_next_edge		; no. 
ir_first_edge				; yes, init ir_t0 and continue
	bsf	ir_state,IR_GOT_FIRST
	bcf	INTCON,GIEH
	movff	TMR1L,ir_t0		; ir_t0 = timer1
	movff	TMR1H,ir_t0+1
	movlw	0x2
	movwf	ir_timeout		; timeout 15~30ms (2x 1/64th sec)
	bcf	ir_state,IR_T1_ROLLOVER
	bsf	INTCON,GIEH
	return
ir_next_edge
	bcf	INTCON,GIEH
	movff	TMR1L,ir_t1		; ir_t1 = timer1
	movff	TMR1H,ir_t1+1

	movff	ir_t1+0,ir_len+0	; compute ir_len = ir_t1 - ir_t0
	movff	ir_t1+1,ir_len+1	
	movf	ir_t0,w
	subwf	ir_len,f		; subtract low byte. C = !borrow
	movf	ir_t0+1,w
	btfss	STATUS,C		; adjust high byte with borrow
	incf	WREG,w
	subwf	ir_len+1,f		; subtract high byte
	movlw	0x2
	btfsc	ir_state,IR_T1_ROLLOVER	; if timer1 has rolled over,
	addwf	ir_len+1,f		; add 0x200 to the result
	movlw	0x2
	movwf	ir_timeout		; timeout 15~30ms (2x 1/64th sec)
	bcf	ir_state,IR_T1_ROLLOVER
	bsf	INTCON,GIEH
	movff	ir_t1+0,ir_t0+0		; start time for next pulse
	movff	ir_t1+1,ir_t0+1
	
	;; start decoding (pulse  measured in ir_t1 (16 bits))
	
	btfss	ir_state,IR_9MS		; need to see first 9ms pulse ?
	bra	check_9ms

	;; at this point, any pulse longer than 7.8ms is bad.
	movf	ir_len+1,w
	bnz	init_ir

	btfss	ir_state,IR_4MS5	; need to see second 4.5ms pulse ?
	bra	check_4ms5

	btfss	ir_state,IR_560US	; got the 560us pulse ?
	bra	check_560us

	;; at this point, we should see a 560us (got a valid 0) or
	;; 3*560us = 2.24ms (got a valid 1)
	movlw	(((.100 - MARGIN) * T_560US) / .100)
	cpfsgt	ir_len
	bra	check_one
	movlw	(((.100 + MARGIN) * T_560US) / .100)
	cpfslt	ir_len
	bra	check_one

	;; got a valid 0
	bcf	STATUS,C
	bra	shift_data

check_one
	movlw	(((.100 - MARGIN) * T_1MS680) / .100)
	cpfsgt	ir_len
	bra	init_ir
	movlw	(((.100 + MARGIN) * T_1MS680) / .100)
	cpfslt	ir_len
	bra	init_ir

	;; got a valid 1
	bsf	STATUS,C
	; bra	shift_data

shift_data
	rrcf	ir_data+3,f		; shift carry in (32 bits)
	rrcf	ir_data+2,f
	rrcf	ir_data+1,f
	rrcf	ir_data+0,f
	bcf	ir_state,IR_560US	; no  => get ready to receive next bit	
	decfsz	ir_bitcount,f		; got 32 bits ?
	return
	;bra	got_data		; yes => check

got_data
	movf	ir_data+3,w		; ir_data[3] = !command
	xorwf	ir_data+2,w		; ir_data[2] = command
	incf	WREG,w			; (!command ^ command) + 1 should be 0
	bnz	init_ir
	movf	ir_data+1,w		; ir_data[1] = !address
	xorwf	ir_data+0,w		; ir_data[0] = address
	incf	WREG,w			; (!address ^ command) + 1 should be 0
	bnz	init_ir
					; data validated
	bsf	GFlag,GOT_IR
	bra	stop_ir


;
; initialize IR
;	
init_ir
	GLOBAL	init_ir
	bsf	IR_DIR			; pin is input
	bcf	IR_INT_IP		; lower priority interrupt
	bcf	IR_INT_IF		; clear eventual pending interrupt
	bsf	IR_INT_IE		; enable interrupt
	bcf	IR_EDGE_REG,IR_EDGE_BIT	; trigger on low edge
	clrf	ir_state		; reset state
	bcf	GFlag,GOT_IR
	return	
	
;
; check that we just got a 9ms pulse. Actually, we cheat and accept a pulse
; longer than 0xFF ticks (that's 7.8ms)
; 
check_9ms
	movf	ir_len+1,w		; if high byte is 0
	bz	init_ir			; then pulse < 7.8ms, reject it.
	bsf	ir_state,IR_9MS
	return

;
; check that we got a 4.5ms pulse. upper limit (7.8ms) already checked
; (note: we enter with W = ir_t1+1)
;
check_4ms5
	movlw	(((.100 - MARGIN) * T_4MS5) / .100)
	cpfsgt	ir_len
	bra	init_ir
	bsf	ir_state,IR_4MS5
	movlw	.32
	movwf	ir_bitcount
	return

;
; check that we got a 560us pulse
;
check_560us
	movlw	(((.100 - MARGIN) * T_560US) / .100)
	cpfsgt	ir_len
	bra	init_ir
	movlw	(((.100 + MARGIN) * T_560US) / .100)
	cpfslt	ir_len
	bra	init_ir
	bsf	ir_state,IR_560US
	return



	
#endif
	
	END
