;;;
;;; Copyright (C) 2010 Frederic Vecoven
;;;
;;; This file is part of NixieClock
;;;
;;; NixieClock is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; NixieClock is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program. If not, see <http://www.gnu.org/licenses/>.
;;;
	
;
; Nixie routines
;
	include hv.inc
	
	; variables
	extern	sfsr, t0fsr
	extern	GFlag, lat_t
	extern	pair0, pair1
	extern	tval, tcount, tcurrent, tmin, tchange
	extern	ttime, ttotal, tloop, tpair, tlist, tdigit, max_bright
#if NUM_TUBES == 6
	extern	pair2
#endif
#if SUPPORT_DP == 1
	extern	pairdp
#endif
	
	; function
	extern	remap_fading, remap_tube
#if REQUIRE_DIGIT_MAPPING == 1
	extern	remap_digit, unmap_digit
#endif

	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; DESIGN 
;
; 4 tubes clock : tubes are numbered  1 2 3 4 
; where tubes 1 & 3 are controlled by ANODE0, and tubes 2 & 4 by ANODE1.
;       tubes 1,2 and 3,4 are connected to two BCD->10 decoders.
;
; 6 tubes clock : tubes are numbered  1 2 3 4 5 6
; where tubes 1 & 4 are controlled by ANODE0, and tubes 2 & 5 by ANODE1,
;          and tubes 3 & 6 by ANODE2
;       tubes 1,2,3 and 4,5,6 are connected to two BCD->10 decoders.
;
; Each tube has a set of parameters :
;  0-9  : primary digit
;  0-9  : secondary digit (for fading)
;  0-31 : fading (0 = full primary digit, 31 = almost-full secondary digit)
;  1-15 : brightness
;
;  To turn off a tube, don't set brightness to 0, but use a digit value > 9.
;
;  These parameters can be updated at any time, and are typically updated
;  every 1/64th second.
;
; 
; Timer0 is used to refresh the tubes. Each pair of tubes get a 4.096ms 
; refresh period. During this period, the primary digit may be changed into
; the secondary (for fading), or the tube can be turned off (to dim brightness).
; This period is split into 256 chunks, so timer0 is used in 8 bits mode.
; Finally, before moving to the next pair of tube, a blanking period
; is inserted.
;
; To minimize overhead, a single timer is used for everything. The algorithm 
; works as follow :
;    1. The primary digit is shown
;    2. The timer is programmed to generate an interrupt when the displayed
;       digits should be changed.
;    3. If we are done with this pair of tubes, we blank (turn off all anodes)
;       and program the timer for the blanking duration.
;    4. Switch to the next pair of tubes, and go to step 1.
;
; To do this, the parameters of each tubes are processed, and a list of events
; is created. Let's use one example (2 pairs of tubes) :
;
; Tube 1 :  primary digit = 1   secondary = 1   fading = 0   brightness = 15
; Tube 2 :  primary digit = 2   secondary = 3   fading = 5   brightness = 13
; Tube 3 :  primary digit = 5   secondary = 5   fading = 0   brightness = 4
; Tube 4 :  primary digit = 9   secondary = 0   fading = 25  brightness = 14
;
; 1) Compute duration of each digit : multiply fading by brightness, divide by 2
;    Add the duration of the first step to the 2nd step
; 
;  Tube1: "1" (32-0)*15 = 480    480 / 2 = 240    t = 0
;         "1"     0 *15 = 0        0 / 2 =   0    t = 240
;                                                 t = 240+0 = 240
;  Tube2: "2" (32-3)*13 = 377    377 / 2 = 188    t = 0
;         "3"     3 *13 = 39      39 / 2 =  19    t = 188
;                                                 t = 188+19 = 207
;  Tube3: "5" (32-0)*4  = 128    128 / 2 =  64    t = 0
;         "5"     0 *4  = 0        0 / 2 =   0    t = 64
;                                                 t = 64+0 = 64
;  Tube4: "9" (32-25)*14= 98      98 / 2 =  49    t = 0
;         "0"     25*14 = 350    350 / 2 = 175    t = 49
;                                                 t = 49+175 = 224
;
; 2) sort the duration per tube pair, and build the raw values to send to 
;    the BCD->10 decoders
;
;  t=0    tubes13="15"         t=0    tubes24="29"
;  t=64   tubes13="1F"         t=49   tubes24="20"
;  t=240  tubes13="FF"         t=188  tubes24="30"
;                              t=207  tubes24="F0"
;                              t=224  tubes24="FF"
;
; 3) make the timing relative
;
;  "15"  t=0                   "29"  t=0
;  "1F"  t=64-0=64             "20"  t=49-0=49
;  "FF"  t=240-64=176          "30"  t=188-49=139
;                              "F0"  t=207-188=19
;                              "FF"  t=224-207=17
;	
; 4) arrange the last data in an array :
;
; "15" 64 "1F" 176 "FF" 0      "29" 49 "20" 139 "30" 19 "F0" 17 "FF" 0
;
; This can be read as : show "15"; after 64 ticks,show "1F"; etc...
;
; 5) terminate the array by the blanking, using a magic value "EE" and
;    the blanking duration.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

nixie_code1	CODE

	
tube_init
	GLOBAL	tube_init
	lfsr	FSR0,tlist
	movff	FSR0L,t0fsr+0
	movff	FSR0H,t0fsr+1
	bcf	GFlag,TUBE_PAIR
	bcf	GFlag,TUBE_PAIR2
	movf	max_bright,w		; set max_bright for all tubes
	movwf	pair0+1	
	movwf	pair0+3
	movwf	pair1+1
	movwf	pair1+3
#if NUM_TUBES == 6
	movwf	pair2+1
	movwf	pair2+3
#endif
	bsf	GFlag,TUBE_UPDATE
	return
	

; return fading of a tube (in 'tcount'). value returned in W
get_fading
	GLOBAL	get_fading
	lfsr	FSR2,pair0
	movf	tcount,w
	call	remap_fading
	movf	PLUSW2,w
	return
	
; set fading value of tube given in 'tcount'. fading in W.
set_fading
	GLOBAL	set_fading
	andlw	0x1F
	movwf	tdigit
	lfsr	FSR2,pair0
	movf	tcount,w
	call	remap_fading
	movff	tdigit,PLUSW2
	return
	
	
; set value of a tube (given in 'tcount'), value in W.
;
; tube1     : tcount == 0
; tube1_alt : tcount == 1
; tube2     : tcount == 2
; tube2_alt : tcount == 3
; ...
set_tube
	GLOBAL	set_tube
	andlw	0x0F
#if REQUIRE_DIGIT_MAPPING == 1
	call	remap_digit
#endif
	movwf	tdigit
	lfsr	FSR2,pair0
	movf	tcount,w
	call	remap_tube
	movff	tdigit,PLUSW2
	return	

; return value of a tube (in 'tcount'). value returned in W
get_tube
	GLOBAL	get_tube
	lfsr	FSR2,pair0
	movf	tcount,w
	call	remap_tube
	movf	PLUSW2,w
	andlw	0x0F
#if REQUIRE_DIGIT_MAPPING == 1
	call	unmap_digit
#endif
	return

	
; return brightness of a tube (in 'tcount'). value returned in W
get_bright
	GLOBAL	get_bright
	lfsr	FSR2,pair0
	movf	tcount,w
	call	remap_fading
	incf	WREG,w
	movf	PLUSW2,w
	return
	
; set brightness of a tube (in 'tcount'), brightness in W
; (if value is lower than BRIGHT_MIN, set BRIGHT_MIN. if higher than 0xF,
; set 0xF).
set_bright
	GLOBAL	set_bright
	movwf	tdigit			; save brightness in tdigit

	cpfsgt	max_bright		; if (max_bright <= tdigit)
	movff	max_bright,tdigit	; then tdigit = max_bright
	
sb_00	movlw	BRIGHT_MIN		; if (tdigit < min_bright)
	cpfsgt	tdigit			; then tdigit = min_bright
	movwf	tdigit
	
	lfsr	FSR2,pair0		; set new brightness
	movf	tcount,w
	call	remap_fading
	incf	WREG,w
	movff	tdigit,PLUSW2
	
	return


; increase max_bright, and wrap over to BRIGHT_MIN
; (called from editor)
inc_max_bright
	GLOBAL	inc_max_bright
	incf	max_bright,w
	andlw	0x0F			; 0xF is the max
	movwf	tdigit			; save in tdigit for now
	movlw	BRIGHT_MIN
	cpfsgt	tdigit			; if (tdigit < min_bright)
	movwf	tdigit			; then tdigit = min_bright
	movff	tdigit,max_bright
	return

	
;
; set decimal point of a tube (given in 'tcount'). value in W
; 	W = 0 : reset dp
; 	W = 1 : set dp
; 	W = 2 : toggle dp
;
; DP values are stored in pairdp[0..1] (4 tubes) or pairdp[0..2] (6 tubes)
;	
; The mapping of a decimal point of a given tube to a bit in pairdp is :
;	
;  tube   tcount   pairdp    bit        tube   tcount   pairdp   bit
; ------+--------+--------+------      ------+--------+--------+-----
;   1   |   0    |   0    |  0           1   |   0    |   0    |  0
;   2   |   1    |   1    |  0           2   |   1    |   1    |  0
;   3   |   2    |   2    |  0           3   |   2    |   0    |  1
;   4   |   3    |   0    |  1           4   |   3    |   1    |  1
;   5   |   4    |   1    |  1
;   6   |   5    |   2    |  1
;
; The 4 tubes case is easy :
;	bit y of pairdp + x = value  where x = bit 0 of tcount
;                                          y = bit 1 of tcount
;
; The 6 tubes cases is a little harder, except if we add 1 to tcount if
; tcount > 3. In this case, tcount becomes 0,1,2,4,5,6 and we have
;
;	bit y of pairdp + x = value  where x = bits 1:0 of tcount
;                                          y = bit 2 of tcount
;
; The 4 tubes case can be seen as the 6 tubes case, by adding 2 to tcount
; if tcount > 1. Tubes 2 and 3 are remapped to tcount positions 4 and 5,
; and the above formula is also valid.
;
; The table (in binary) with tcount tweaked, is the following :	
;
; tcount'  pairdp  bit
; -------+-------+------
;  0 0 0 |  0 0  |  0
;  0 0 1 |  0 1  |  0
;  0 1 0 |  1 0  |  0
;  0 1 1 |  x x  |  x
;  1 0 0 |  0 0  |  1
;  1 0 1 |  0 1  |  1
;  1 1 0 |  1 0  |  1
;  1 1 1 |  x x  |  x
;
#if SUPPORT_DP == 1
set_dp
	GLOBAL	set_dp
	movwf	tdigit			; save value (1 = set dp, 0 = reset dp)
	movf	tcount,w
#if NUM_TUBES == 6
	addlw	-.3
	btfsc	STATUS,C
	addlw	.1
	addlw	.3
#else
	addlw	-.2
	btfsc	STATUS,C
	addlw	.2
	addlw	.2
#endif
	movwf	tdigit+1		; tdigit+1 = tweaked tcount
	lfsr	FSR2,pairdp
	andlw	0x3
	addwf	FSR2L,f			; INDF2 = pairdp[tdigit+1[1:0]]

	btfsc	tdigit,1		; toggle ?
	bra	set_dp_toggle
	btfss	tdigit,0		; set or reset ?
	bra	set_dp_0
set_dp_1
	movlw	0x1
	btfsc	tdigit+1,2		; set bit 0 or bit 1 of INDF2 ?
	movlw	0x2
	iorwf	INDF2,f	
	return
set_dp_0
	movlw	0xFE
	btfsc	tdigit+1,2		; reset bit 0 or bit 1 of INDF2 ?
	movlw	0xFD
	andwf	INDF2,f
	return
set_dp_toggle
	movlw	0x1
	btfsc	tdigit+1,2		; toggle bit 0 or bit 1 of INDF2 ?
	movlw	0x2
	xorwf	INDF2,f
	return
#endif
	
	

	
	
;
; Rebuild the array containing the tube state changes.
;
; "pair data" 
;      FADE1  (0-31)  BRIGHTNESS1  (0-16)  FADE2 (0-31)    BRIGHTNESS2 (0-16)
;      DIGIT1  (0-9)  DIGIT2  (0-9)        DIGIT1' (0-9)   DIGIT2' (0-9)
;
; blanking magic value : 
BLANKING EQU 0xEE
	
rebuild_tlist
	GLOBAL	rebuild_tlist
	movff	FSR1L,sfsr+2
	movff	FSR1H,sfsr+3
	movff	FSR2L,sfsr+4
	movff	FSR2H,sfsr+5
	lfsr	FSR1,tlist		; clear tlist
	movff	FSR1L,tpair		; reset tpair (offset in tlist)
	movlw	NUM_TUBES*6
rt_00	clrf	POSTINC1
	decfsz	WREG,f
	bra	rt_00
	lfsr	FSR1,pair0		; process pair 0
	call	rebuild_pair
	lfsr	FSR1,pair1
	call	rebuild_pair		; process pair 1
#if NUM_TUBES == 6
	lfsr	FSR1,pair2
	call	rebuild_pair		; process pair 2
#endif
	call	shift_tlist
	bcf	GFlag,TUBE_UPDATE
	movff	sfsr+2,FSR1L
	movff	sfsr+3,FSR1H
	movff	sfsr+4,FSR2L
	movff	sfsr+5,FSR2H
	return
	
rebuild_pair
	lfsr	FSR2,tval
	movlw	.2
	movwf	tloop
rp_0	movlw	.32
	bsf	STATUS,C
	subfwb	POSTINC1,w
	mulwf	INDF1			; PROD = (32 - fade) * brightness
	rrcf	PRODH,w
	rrcf	PRODL,w
	movwf	POSTINC2		; tval[0] = PROD / 2
	swapf	POSTINC1,w
	movwf	POSTINC2		; tval[1] = 16 * brightness
	decfsz	tloop,f
	bra	rp_0
		
	lfsr	FSR0,tchange
	swapf	POSTINC1,w
	iorwf	POSTINC1,w
	movwf	POSTINC0		; tchange[0] = (digit1 << 4) | digit1'
	clrf	POSTINC0		; ttime[1] = 0
	
	clrf	tcurrent		; tcurrent = 0
	movlw	.4
	movwf	tloop			; for (i = 0; i < 4; i++) {
rp_00	movlw	0xff
	movwf	tmin			;    tmin = 0xff
	lfsr	FSR2,tval
	clrf	tloop+1			;    for (j = 0; j < 4; j++) {
rp_01	movf	tmin,w
	cpfslt	INDF2			;          if (tval[j] < tmin
	bra	rp_02
	movf	tcurrent,w
	subwf	INDF2,w			;          && tval[j] >= tcurrent)
	bnc	rp_02
	movff	tloop+1,tmin+1		;               tmin2 = j   
	movff	INDF2,tmin		;               tmin = tval[j]
rp_02	incf	FSR2L,f		
	incf	tloop+1,f
	movlw	.3
	cpfsgt	tloop+1			;    }
	bra	rp_01
	movff	tmin,tcurrent		;    tcurrent = tmin
	lfsr	FSR2,tval
	movf	tmin+1,w
	clrf	PLUSW2			;    tval[tmin2] = 0

	clrf	WREG
	btfsc	tmin+1,1
	incf	WREG,w
	movff	PLUSW1,tloop+1		;    digit = pair[(tmin2 & 0x2) ? 7: 6]
	btfss	tmin+1,0
	bra	rp_03
	movlw	0xF
	movwf	tloop+1			;    digit = (tmin2 & 0x1) ? 0xF : digit
rp_03	movlw	-.2
	movf	PLUSW0,w
	movwf	POSTINC0		;    tchange[i+1] = tchange[i]
	movff	tmin,POSTDEC0		;    ttime[i+1] = tmin
	movf	INDF0,w
	btfss	tmin+1,1		;    if (tmin2 & 0x2) {
	bra	rp_04
	andlw	0xF0			;        tchange[i+1] &= 0xF0
	iorwf	tloop+1,w		;        tchange[i+1] |= digit
	bra	rp_05			;    } else {
rp_04	andlw	0x0F			;        tchange[i+1] &= 0x0F
	swapf	tloop+1,f
	iorwf	tloop+1,w		;        tchange[i+1] |= (digit << 4)
rp_05	movwf	POSTINC0
	incf	FSR0L,f
	
	decfsz	tloop,f
	bra	rp_00			; }
	
	lfsr	FSR0,tlist
	movff	tpair,FSR0L
	lfsr	FSR1,tchange
	movff	POSTINC1,POSTINC0	; tlist[0] = tchange[0]
	movf	POSTINC1,w
	movwf	POSTINC0		; tlist[1] = ttime[0]
	movwf	ttime
	clrf	ttotal
	movlw	.4
	movwf	tloop

rp_09
	movff	POSTINC1,tmin		; tmin = tchange[i]
	movf	INDF1,w			; w = ttime[i]
	cpfseq	ttime			; if (ttime[i-1] == ttime[i])
	bra	rp_10			; then
	decf	FSR0L,f
	decf	FSR0L,f
	movff	tmin,POSTINC0		;    tlist[j-2] = tchange[i]
	incf	FSR0L,f
	bra	rp_11
rp_10					; else
	movff	tmin,POSTINC0		;    tlist[j++] = tchange[i]
	bsf	STATUS,C
	subfwb	ttime,w	
	movwf	INDF0			;    tlist[j] = ttime[i] - ttime[i-1]
	movf	ttotal,w
	addwf	POSTINC0,w
	movwf	ttotal			;    ttotal += tlist[j++]
rp_11
	movff	POSTINC1,ttime
	decfsz	tloop,f
	bra	rp_09

	movlw	BLANKING
	movwf	POSTINC0		; tlist[j++] = BLANKING
	movlw	.240 			; MAXTIME
	bsf	STATUS,C
	subfwb	ttotal,w
	movwf	POSTINC0		; tlist[j++] = MAXTIME - ttotal
	movff	FSR0L,tpair		; save offset in tlist
	
	return

;
; the value of tlist[1] appears in tlist[3], and so on. shift these values
; to the correct positions
;	
shift_tlist
	lfsr	FSR0,tlist+1
	lfsr	FSR1,tlist+3
	movlw	NUM_TUBES*3-1
	movwf	tloop
st_00	movff	POSTINC1,POSTINC0
	movf	POSTINC0,w
	movf	POSTINC1,w
	decfsz	tloop,f
	bra	st_00
	return


;
; timer 0 : refresh the tubes. Set a value, and reprogram timer 0
;           to trigger when a different value should be displayed.
;
int_timer0
	GLOBAL	int_timer0
	bcf	INTCON,TMR0IF

	movff	FSR0L,sfsr+0		; save FSR0
	movff	FSR0H,sfsr+1
	movff	t0fsr+0,FSR0L		; restore our FSR0.
	movff	t0fsr+1,FSR0H

#if NUM_TUBES == 6
	btfsc	GFlag,TUBE_PAIR2
	bra	refresh_pair_2
#endif
	btfsc	GFlag,TUBE_PAIR
	bra	refresh_pair_1
refresh_pair_0
	bcf	ANODE1			; get data from list, and display
#if NUM_TUBES == 6
	bcf	ANODE2
#endif
	call	set_data
#if SUPPORT_DP == 1
	movf	pairdp+0,w
	call	set_data_dp
#endif
	bsf	ANODE0
	movf	POSTINC0,w		; get timing. 0 = no delay
	bz	zero_pair_0
	negf	WREG
	movwf	TMR0L			; not 0 = timer delay
	bra	t0_done
zero_pair_0
	movf	INDF0,w			; look at next data. 
	sublw	BLANKING		; blanking ?
	bz	refresh_pair_0		; yes => skip this update
	bcf	ANODE0			; no => blanking period
	bcf	ANODE1
#if NUM_TUBES == 6
	bcf	ANODE2
#endif	
	movlw	.250
	movwf	TMR0L
	bsf	GFlag,TUBE_PAIR		; next pair is pair1
	bra	t0_done
refresh_pair_1
	bcf	ANODE0			; get data from list, and display
#if NUM_TUBES == 6
	bcf	ANODE2
#endif	
	call	set_data
#if SUPPORT_DP == 1
	movf	pairdp+1,w
	call	set_data_dp
#endif	
	bsf	ANODE1
	movf	POSTINC0,w		; get timing. 0 = no delay
	bz	zero_pair_1
	negf	WREG
	movwf	TMR0L			; not 0 = timer delay
	bra	t0_done
zero_pair_1
	movf	INDF0,w			; look at next data. 
	sublw	BLANKING		; blanking ?
	bz	refresh_pair_1		; yes => skip this update
	bcf	ANODE0			; no => blanking period
	bcf	ANODE1
#if NUM_TUBES == 6
	bcf	ANODE2
#endif	
	movlw	.250
	movwf	TMR0L
	bcf	GFlag,TUBE_PAIR		; next pair is pair0 (4 tubes)
#if NUM_TUBES == 6
	bsf	GFlag,TUBE_PAIR2	; or pair2 (6 tubes)
	bra	t0_done
refresh_pair_2
	bcf	ANODE0			; get data from list, and display
	bcf	ANODE1
	call	set_data
#if SUPPORT_DP == 1
	movf	pairdp+2,w
	call	set_data_dp
#endif	
	bsf	ANODE2
	movf	POSTINC0,w		; get timing. 0 = no delay
	bz	zero_pair_2
	negf	WREG
	movwf	TMR0L			; not 0 = timer delay
	bra	t0_done
zero_pair_2
	movf	INDF0,w			; look at next data
	sublw	BLANKING		; blanking ?
	bz	refresh_pair_2		; yes => skip this update
	bcf	ANODE0			; no => blanking period
	bcf	ANODE1
	bcf	ANODE2
	movlw	.250
	movwf	TMR0L
	bcf	GFlag,TUBE_PAIR2	; next pair is pair0 (6 tubes)
#endif	
	btfsc	GFlag,TUBE_UPDATE	; rebuild tlist if necessary
	call	rebuild_tlist	
	lfsr	FSR0,tlist		; reset FSR0 (back to pair0)
t0_done
	movff	FSR0L,t0fsr+0		; save our FSR0
	movff	FSR0H,t0fsr+1
	movff	sfsr+0,FSR0L		; and restore it.
	movff	sfsr+1,FSR0H
	return	

; set 8 bits (in INDF0, return with FSR0++) -> tubes
set_data
	GLOBAL	set_data
			; 4 MSB -> PORTB[7..4]
	movf	LATB,w		
	andlw	0x0F
	movwf	lat_t
	swapf	INDF0,w
	andlw	0xF0
	iorwf	lat_t,w
	movwf	LATB
			; 4 LSB -> PORTC[7..4]
	movf	LATC,w
	andlw	0x0F
	movwf	lat_t
	movf	POSTINC0,w
	andlw	0xF0
	iorwf	lat_t,w
	movwf	LATC
	return

;
; Set the dots cathode controller
;
; We cannot just set the DP gpio based on the dot on/off value, because
; by doing so, the DP will be full brightness. When the dot value is ON,
; we check that the tubes are actually displaying something (we just check
; the BCD->decimal converter is not driven with binary 11xx)
;
#if SUPPORT_DP == 1
set_data_dp
	btfss	WREG,0			; dot on ?
	bra	dp0_off			; no ==> off
	btfss	LATB,LATB_D_BIT		; yes ==> check for 11xx
	bra	dp0_on
	btfsc	LATB,LATB_C_BIT
	bra	dp0_off			; C and D bits are set => DP off.
dp0_on
	bsf	DP0
	bra	set_dp1
dp0_off
	bcf	DP0
set_dp1
	btfss	WREG,1
	bra	dp1_off
	btfss	LATC,LATC_D_BIT
	bra	dp1_on
	btfsc	LATC,LATC_C_BIT
	bra	dp1_off
dp1_on
	bsf	DP1
	return
dp1_off
	bcf	DP1
	return
#endif	
	
	END
	